# -*- coding: utf-8 -*-
"""
Payment History service, records modification done on payments
"""
import logging
from endi_payment.models import EndiPaymentHistory


class HistoryLogService(object):
    """
    History service that's only logging the actions in a log file.

    To be used in dev mode
    """
    def __init__(self, context, request):
        self.request = request
        self.logger = logging.getLogger("endi_payment")

    def record(self, action, invoice, payment):
        self.logger.debug("Recording an action")
        self.logger.debug("Action : %s" % action)
        self.logger.debug("User : %s" % self.request.user)
        self.logger.debug("Invoice id : %s" % invoice.id)


class HistoryDBService(HistoryLogService):
    """
    Service logging datas in a database using a specific connexion
    """

    def record(self, action, invoice, payment):
        self.logger.info("Recording an action")
        self.logger.info("Action : %s" % action)
        self.logger.info("User : %s" % self.request.user)
        self.logger.info("Invoice id : %s" % invoice.id)
        self.logger.info("It's production mode")

        from endi_payment.database import LocalSessionContext
        with LocalSessionContext() as session:
            record = EndiPaymentHistory(
                action_type=action,
                payment_id=payment.id,
                mode=payment.mode,
                amount=payment.amount,
                bank_remittance_id=payment.bank_remittance_id,
                date=payment.date,
                invoice_id=invoice.id,
                invoice_pdf_file_hash=invoice.pdf_file_hash,
                bank_cg=payment.bank.compte_cg,
                tva_value=payment.tva.value,
                user_login=self.request.user.login.login,
            )
            session.add(record)
        self.logger.info(u"Recorded")



# -*- coding:utf-8 -*-
from pyramid.path import DottedNameResolver

from endi_payment.interfaces import IPaymentRecordHistoryService


def includeme(config):
    settings = config.get_settings()

    module_path = "endi_payment.history.HistoryLogService"

    key = "endi_payment.interfaces.IPaymentRecordHistoryService"
    if key in settings:
        module_path = settings[key]

    if module_path == "endi_payment.history.HistoryDBService":
        if 'endi_payment_db.url' not in settings:
            raise Exception(u"endi_payment.dev est à false et aucune url "
                            u"endi_payment_db.url n'est fourni")
        else:
            # We will store the endi payment's history
            config.include('.database')

    history_service = DottedNameResolver().resolve(module_path)

    config.register_service_factory(
        history_service, IPaymentRecordHistoryService
    )



# -*- coding: utf-8 -*-
from zope.interface import Interface


class IPaymentRecordHistoryService(Interface):
    """
    History manipulation tool iinterface for Payment action log
    """

    def record_action(self, action, invoice, payment):
        """
        History manipulation tool for Payment registration
        """
        pass



# -*- coding: utf-8 -*-
import datetime
from sqlalchemy import (
    Column,
    Integer,
    BigInteger,
    String,
    DateTime,
)
from autonomie_base.models.base import default_table_args
from endi_payment.database import ModelBase


class EndiPaymentHistory(ModelBase):
    __tablename__ = "endi_payment_history"
    __table_args__ = default_table_args
    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    # ADD/UPDATE/DELETE
    action_type = Column(String(6))
    # Original payment instance id
    payment_id = Column(Integer, nullable=False)
    # Payment mode
    mode = Column(String(50))
    # The amount of the payment (* 10 000)
    amount = Column(BigInteger(), nullable=False)
    # The string identifying the bank remittance id
    bank_remittance_id = Column(String(255))
    # Date the payment has been received
    date = Column(DateTime(), nullable=False)
    # The invoice database item's id
    invoice_id = Column(Integer, nullable=False)
    # The invoice PDF sha1 hash that can be used to check pdf files
    invoice_pdf_file_hash = Column(String(255), nullable=False)
    # The bank CG account
    bank_cg = Column(String(120), default="")
    # The TVA associated to the recorded payment
    tva_value = Column(Integer, nullable=False)
    # The login of the user responsible for this modification
    user_login = Column(String(64), nullable=False)



# -*- coding: utf-8 -*-
"""
Specific session stuff used to store the payment history logs
"""
from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

HistorySessionFactory = sessionmaker()
ModelBase = declarative_base()


class CriticalIntegrityError(Exception):
    """
    Error raised when we can't ensure the integrity of the current managed datas
    anymore
    """
    pass


class LocalSessionContext(object):
    def __enter__(self):
        """
        Initialize a specific session
        """
        self.session = HistorySessionFactory()
        return self.session

    def __exit__(self, type, value, traceback):
        try:
            self.session.commit()
        except:
            self.session.rollback()
            raise CriticalIntegrityError(
                u"Erreur critique rencontrée lors de la journalisation "
                u"des évènements d'encaissement"
            )


def configure_specific_payment_engine(config):
    """
    Configure un bindind spécifique (avec un utilisateur différent) pour la
    manipulation des instances de Payment

    Garder pour le principe mais ne sera peut être jamais utilisé.
    """
    settings = config.get_settings()
    if 'endi_payment_db.url' in settings:
        from sqlalchemy.orm import Session
        prefix = 'endi_payment_db.'

        engine = engine_from_config(settings, prefix=prefix)
        main_engine = engine_from_config(settings)

        class CustomSession(Session):
            def __init__(self, *args, **kwargs):
                Session.__init__(self, *args, **kwargs)

            def get_bind(self, mapper=None, clause=None):
                from autonomie.models.task.invoice import Payment
                if mapper is not None and issubclass(mapper.class_, Payment):
                    return engine
                else:
                    return main_engine

        from autonomie_base.models.base import DBSESSION
        DBSESSION.configure(class_=CustomSession)


def includeme(config):
    """
    Pyramid Include's mechanism
    Setup the library specific session
    """
    settings = config.get_settings()

    if 'endi_payment_db.url' in settings:
        # On a une connexion particulière pour l'édition des journaux
        prefix = 'endi_payment_db.'
        endi_payment_engine = engine_from_config(settings, prefix=prefix)

    else:
        # On utilise l'engine sqlalchemy par défaut (celui d'endi)
        endi_payment_engine = endi_payment_engine(settings)

    from endi_payment.models import EndiPaymentHistory  # NOQA

    HistorySessionFactory.configure(bind=endi_payment_engine)
    ModelBase.metadata.bind = endi_payment_engine
    ModelBase.metadata.create_all(endi_payment_engine)

    from transaction import commit, begin
    commit()
    begin()
    return True



# -*- coding: utf-8 -*-
"""
Public Tools are defined here
"""
from autonomie.models.task.invoice import Payment
from endi_payment.interfaces import IPaymentRecordHistoryService


class PaymentService(object):
    """
    A class grouping all the public methods

    record_payment

        Record a new payment

    modifiy_payment

        Modify an existing payment


    delete_payment

        Delete an existing payment
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.dbsession = request.dbsession
        self.history = request.find_service(IPaymentRecordHistoryService)

    def add(self, invoice, params):
        """
        Record a payment for the given invoice

        :param obj invoice: The Invoice associated to this transaction
        :param **params: List of parameters used to record a payment

        :returns: The generated Payment object id
        :rtype: int
        """
        payment = Payment()
        for key, value in params.items():
            setattr(payment, key, value)
        payment.user_id = self.request.user.id
        invoice.payments.append(payment)
        self.dbsession.merge(invoice)
        self.dbsession.flush()
        self.history.record(
            "ADD",
            invoice,
            payment,
        )
        return payment

    def update(self, payment, params):
        """
        Modify the given Payment instance
        :param obj invoice: The Invoice associated to this transaction
        :param **params: List of parameters used to record a payment

        :returns: The modified Payment object id
        :rtype: int
        """
        for key, value in params.items():
            setattr(payment, key, value)
        payment = self.dbsession.merge(payment)

        self.history.record(
            "UPDATE",
            payment.task,
            payment,
        )
        return payment

    def delete(self, payment):
        """
        Delete the given Payment instance

        :param obj invoice: The Invoice associated to this transaction

        :returns: True/False if the deletion succeeded
        :rtype: bool
        """
        self.dbsession.delete(payment)
        self.dbsession.flush()
        self.history.record(
            "DELETE",
            payment.task,
            payment,
        )
        return True
